#include <stdio.h>
#include <stdlib.h>
#include <string.h>


char * patch = NULL;

#define STR_LINE_MAX 512
char str_line[STR_LINE_MAX];

typedef enum {
   LT_ERROR,
   LT_COMMENT,
   LT_FILE,
   LT_BASE,
   LT_OFFSET,
   LT_BYTES,
   LT_STRING
} LINE_TYPE_E;


// ==========================================================
char * get_str_line(char * cptr, LINE_TYPE_E * line_type)
{
   long i = 0, max = STR_LINE_MAX - 1;
   char c;


   c = * cptr;
   while ((c != 0x00) && (c != 0x0D) && (c != 0x0A) && (i < max))
   {
      str_line[i] = c;
      i++;
      cptr++;
      c = * cptr;
   }
   str_line[i] = 0x00;

   // skip new lines
   while ((c != 0x00) && ((c == 0x0D) || (c == 0x0A)))
   {
      cptr++;
      c = * cptr;
   }

   // line type
   * line_type = LT_ERROR;

   if (str_line[0] == ';')
      * line_type = LT_COMMENT;
   else if ((strnicmp(str_line, "file", 4) == 0) && ((str_line[4] == '=') || (str_line[4] == ' ')))
      * line_type = LT_FILE;
   else if ((strnicmp(str_line, "base", 4) == 0) && ((str_line[4] == '=') || (str_line[4] == ' ')))
      * line_type = LT_BASE;
   else if ((strnicmp(str_line, "offset", 6) == 0) && ((str_line[6] == '=') || (str_line[6] == ' ')))
      * line_type = LT_OFFSET;
   else if ((strnicmp(str_line, "bytes", 5) == 0) && ((str_line[5] == '=') || (str_line[5] == ' ')))
      * line_type = LT_BYTES;
   else if ((strnicmp(str_line, "string", 6) == 0) && ((str_line[6] == '=') || (str_line[6] == ' ')))
      * line_type = LT_STRING;

   return cptr;
}


// ==========================================================
int main(int argc, char ** argv)
{
   FILE          * in;
   long          size, base = 0, offset, pos;
   int           ret, i;
   char          * cptr, c;
   unsigned char byte, * bptr;
   LINE_TYPE_E   line_type;


   printf(
      "paul_patch v 1.00 freeware\n"
      "==========================\n\n"
   );

   if (argc != 2)
   {
      printf("syntaxe : paul_patch <file.txt>\n");
      return 0;
   }

   in = fopen(argv[1], "rb");
   if (in == NULL)
   {
      printf("ERROR, can't open \"%s\"\n", argv[1]);
      return -1;
   }
   printf("patch file : \"%s\"\n", argv[1]);

   //copy patch file in memory
   fseek(in, 0, SEEK_END);
   size = ftell(in);
   fseek(in, 0, SEEK_SET);
   patch = (char *) malloc(size + 1);
   if (patch == NULL)
   {
      fclose(in);
      printf("ERROR, can't allocate %li bytes for the copy of patch in memory\n", size + 1);
      return -1;
   }
   ret = fread(patch, size, 1, in);
   if (ret != 1)
   {
      free(patch);
      fclose(in);
      printf("ERROR : can't read %li bytes from patch\n", size);
      printf("ret = %i\n", ret);
      return -1;
   }
   patch[size] = 0x00;
   fclose(in);
   in = NULL;

   // for all lines, one by one
   cptr = patch;
   c    = * cptr;
   while (c != 0x00)
   {
      cptr = get_str_line(cptr, & line_type);
      if (line_type == LT_ERROR)
      {
         printf("ERROR : %s\n", str_line);
      }

      i = 4;
      if ((line_type == LT_OFFSET) || (line_type == LT_STRING))
         i = 6;
      else if (line_type == LT_BYTES)
         i = 5;
      
      while ((str_line[i] != 0x00) && (str_line[i] == ' '))
            i++;
      if (str_line[i] != 0x00)
         i++;
      while ((str_line[i] != 0x00) && (str_line[i] == ' '))
         i++;

      switch (line_type)
      {
         case LT_FILE :
            in = fopen(str_line + i, "rb+");
            if (in == NULL)
            {
               free(patch);
               printf("ERROR : can't open file \"%s\"\n", str_line + i);
               return -1;
            }
            printf("use FILE : %s\n", str_line + i);
            break;

         case LT_BASE :
            if (in == NULL)
            {
               free(patch);
               printf("ERROR : can't use BASE \"%s\", since no file is open\n", str_line + i);
               return -1;
            }
            sscanf(str_line + i, "0x%lX", & base);
            printf("use BASE : 0x%08lX\n", base);
            break;

         case LT_OFFSET :
            if (in == NULL)
            {
               free(patch);
               printf("ERROR : can't go to OFFSET \"%s\", since no file is open\n", str_line + i);
               return -1;
            }
            sscanf(str_line + i, "0x%lX", & offset);
            pos = offset - base + 0x1000;
            if (fseek(in, pos, SEEK_SET) != 0)
            {
               free(patch);
               printf("ERROR : can't fseek() to position %li\n", pos);
               return -1;
            }
            printf("fseek() to OFFSET : 0x%08lX (file byte position : 0x%08lX)\n", offset, pos);
            break;

         case LT_BYTES :
            if (in == NULL)
            {
               free(patch);
               printf("ERROR : can't write bytes since no file is open\n");
               return -1;
            }
            c = str_line[i];
            while (c != 0x00)
            {
               if (sscanf(str_line + i, "%02X", & byte) == 1)
               {
                  if (fwrite( & byte, 1, 1, in) != 1)
                  {
                     free(patch);
                     printf("ERROR : can't write 1 byte\n");
                     return -1;
                  }
                  printf("write BYTE : 0x%02X\n", byte);

                  i += 2;
                  if (str_line[i] == ' ')
                     i++;
                  c = str_line[i];                  
               }
               else
                  c = 0x00;
            }
            fflush(in);
            break;

         case LT_STRING :
            if (in == NULL)
            {
               free(patch);
               printf("ERROR : can't write string since no file is open\n");
               return -1;
            }
            bptr = (unsigned char *) (str_line + i);
            byte = * bptr;
            while (byte != 0x00)
            {
               if (fwrite( & byte, 1, 1, in) != 1)
               {
                  free(patch);
                  printf("ERROR : can't write 1 byte of a string\n");
                  return -1;
               }
               bptr++;
               byte = * bptr;
            }
            printf("write STRING : %s\n", str_line + i);
            fflush(in);
            break;
      }
      c = * cptr;
   }
   
   

   // end
   if (in != NULL)
      fclose(in);
   free(patch);
   return 0;
}
